{% if security.registerable or security.recoverable or security.confirmable or security.changeable %}
<ul>
  {% if security.registerable %}
  <li><a href="{{ url_for_security('register') }}{% if 'next' in request.args %}?next={{ request.args.next|urlencode }}{% endif %}">{{ _('Register') }}</a><br/></li>
  {% endif %}
  {% if security.recoverable %}
  <li><a href="{{ url_for_security('forgot_password') }}">{{ _('Forgot password') }}</a><br/></li>
  {% endif %}
  {% if security.confirmable %}
  <li><a href="{{ url_for_security('send_confirmation') }}">{{ _('Confirm account') }}</a></li>
  {% endif %}
  {% if security.changeable %}
  <li><a href="{{ url_for_security('change_password') }}">{{ _('Change password') }}</a></li>
  {% endif %}
</ul>
{% endif %}
