{% extends "security/base.html" %}
{% from "security/_macros.html" import render_field_with_errors, render_field, header %}

{% block content %}
{% include "security/_messages.html" %}
{{ header(_('Reset password')) }}
<form action="{{ url_for_security('reset_password', token=reset_password_token) }}" method="POST" name="reset_password_form">
  {{ reset_password_form.hidden_tag() }}
  {{ render_field_with_errors(reset_password_form.password) }}
  {{ render_field_with_errors(reset_password_form.password_confirm) }}
  {{ render_field(reset_password_form.submit, class_="button is-link") }}
</form>
{% endblock %}
